#include "UpdateResource.h"
#include "RsrcStructs.h"

#pragma warning(disable : 4786)

#include <vector>
#include <map>
#include <algorithm>
#include <time.h>

using namespace std;

////////////////////////////////////////////////////////////////////////
// Misc. helpers

bool Resource::operator<(const Resource &rc) const {
	//A string is less than an ID.
	//Compare type first, then name and language last.

	if (HIWORD(TypeID)) {
		if (HIWORD(rc.TypeID)) {
			int result = wcsicmp(TypeString, rc.TypeString);
			if (result < 0)
				return true;
			else if (result > 0)
				return false;
		}
		else
			return true;
	}
	else {
		if (HIWORD(rc.TypeID)) {
			return false;
		}
		else {
			if (TypeID < rc.TypeID)
				return true;
			if (TypeID > rc.TypeID)
				return false;
		}
	}

	if (HIWORD(NameID)) {
		if (HIWORD(rc.NameID)) {
			int result = wcsicmp(NameString, rc.NameString);
			if (result < 0)
				return true;
			else if (result > 0)
				return false;
		}
		else
			return true;
	}
	else {
		if (HIWORD(rc.NameID)) {
			return false;
		}
		else {
			if (NameID < rc.NameID)
				return true;
			if (NameID > rc.NameID)
				return false;
		}
	}

	return LanguageID < rc.LanguageID;
}

bool Resource::operator==(const Resource &rc) const {
	if (HIWORD(TypeID)) {
		if (!HIWORD(rc.TypeID))
			return false;

		if (wcsicmp(TypeString, rc.TypeString) != 0)
			return false;
	}
	else {
		if (HIWORD(rc.TypeID))
			return false;

		if (TypeID != rc.TypeID)
			return false;
	}

	if (HIWORD(NameID)) {
		if (!HIWORD(rc.NameID))
			return false;

		if (wcsicmp(NameString, rc.NameString) != 0)
			return false;
	}
	else {
		if (HIWORD(rc.NameID))
			return false;

		if (NameID != rc.NameID)
			return false;
	}

	return LanguageID == rc.LanguageID;
}

static bool RsrcIDEqual(DWORD id1, DWORD id2) {
	if (HIWORD(id1)) {
		if (!HIWORD(id2))
			return false;

		if (wcsicmp((LPWSTR)id1, (LPWSTR)id2) != 0)
			return false;
	}
	else {
		if (HIWORD(id2))
			return false;

		if (id1 != id2)
			return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////
// _BeginUpdateResource()

static LPWSTR GetRsrcString(RsrcInfo *pInfo, char *pRsrc, int Offset) {
	WORD Size = *(WORD *)(pRsrc + Offset);

	for (vector<LPWSTR>::iterator I = pInfo->vtrStrings.begin(); I != pInfo->vtrStrings.end(); I++) {
		if ((wcslen(*I) == Size) && (wcsnicmp((LPWSTR)(pRsrc + Offset + 2), *I, Size) == 0))
			return *I;
	}

	LPWSTR lpStr = new WCHAR[Size + 1];

	memcpy(lpStr, pRsrc + Offset + 2, Size * 2);
	lpStr[Size] = 0;	//null-terminate

	pInfo->vtrStrings.push_back(lpStr);

	return lpStr;
}

static void FillRsrcEntries(RsrcInfo *pInfo, char *pRsrc) {
	DWORD dwType  = 0, dwName = 0;
	WORD  wLangID = 0;

	char *p = pRsrc;

	TableHead Head;

	RsrcDirTableHead *pTypeHead = (RsrcDirTableHead *)p;
	p += sizeof RsrcDirTableHead;

	Head.Level = 0;
	Head.Characteristics = pTypeHead->Characteristics;
	Head.TimeDateStamp = pTypeHead->TimeDateStamp;
	Head.VersionMajor = pTypeHead->VersionMajor;
	Head.VersionMinor = pTypeHead->VersionMinor;

	pInfo->vtrTableHeads.push_back(Head);

	for (int TypeIndex = 0; TypeIndex < pTypeHead->NumNameEntries; TypeIndex++) {	//enumerate resource types
		RsrcDirEntry *pTypeEntry = (RsrcDirEntry *)p;
		p += sizeof(RsrcDirEntry);

		if (pTypeEntry->DataEntryRVA >> 31) {	//pTypeEntry->DataEntryRVA & 0x80000000
			dwType = (DWORD)GetRsrcString(pInfo, pRsrc, pTypeEntry->NameRVA & ~(1<<31));

			char *p2 = pRsrc + (pTypeEntry->SubdirectoryRVA & ~0x80000000);
			RsrcDirTableHead *pNameHead = (RsrcDirTableHead *)p2;
			p2 += sizeof RsrcDirTableHead;

			Head.Level   = 1;
			Head.TypeID  = dwType;
			Head.Characteristics = pNameHead->Characteristics;
			Head.TimeDateStamp   = pNameHead->TimeDateStamp;
			Head.VersionMajor    = pNameHead->VersionMajor;
			Head.VersionMinor    = pNameHead->VersionMinor;

			pInfo->vtrTableHeads.push_back(Head);

			for (int NameIndex = 0; NameIndex < pNameHead->NumNameEntries; NameIndex++) {
				RsrcDirEntry *pNameEntry = (RsrcDirEntry *)p2;
				p2 += sizeof(RsrcDirEntry);

				if (pNameEntry->DataEntryRVA >> 31) {
					dwName = (DWORD)GetRsrcString(pInfo, pRsrc, pNameEntry->NameRVA & ~(1<<31));

					char *p3 = pRsrc + (pNameEntry->SubdirectoryRVA & ~0x80000000);
					RsrcDirTableHead *pLangHead = (RsrcDirTableHead *)p3;
					p3 += sizeof RsrcDirTableHead;

					Head.Level   = 2;
					Head.TypeID  = dwType;
					Head.NameID  = dwName;
					Head.Characteristics = pLangHead->Characteristics;
					Head.TimeDateStamp   = pLangHead->TimeDateStamp;
					Head.VersionMajor    = pLangHead->VersionMajor;
					Head.VersionMinor    = pLangHead->VersionMinor;

					pInfo->vtrTableHeads.push_back(Head);

					for (int LangIndex = 0; LangIndex < pLangHead->NumIDEntries; LangIndex++) {
						RsrcDirEntry *pLangEntry = (RsrcDirEntry *)p3;
						p3 += sizeof(RsrcDirEntry);

						//here can not be subdirs
						wLangID = pLangEntry->IntegerID;

						RsrcDataEntry *pData = (RsrcDataEntry *)(pRsrc + (pLangEntry->DataEntryRVA & ~0x80000000));

						Resource rc;

						rc.NameID     = dwName;
						rc.TypeID     = dwType;
						rc.LanguageID = wLangID;
						rc.DataSize   = pData->Size;
						rc.Codepage   = pData->Codepage;
						rc.Reserved   = pData->Reserved;

						rc.Data = new BYTE[pData->Size];
						memcpy(rc.Data, pRsrc + pData->DataRVA - pInfo->SectionHeader.VirtualAddress, pData->Size);

						pInfo->vtrResources.push_back(rc);
					}
				}
			}

			for (NameIndex = 0; NameIndex < pNameHead->NumIDEntries; NameIndex++) {
				RsrcDirEntry *pNameEntry = (RsrcDirEntry *)p2;
				p2 += sizeof(RsrcDirEntry);

				if (pNameEntry->DataEntryRVA >> 31) {
					dwName = pNameEntry->IntegerID;

					char *p3 = pRsrc + (pNameEntry->SubdirectoryRVA & ~0x80000000);
					RsrcDirTableHead *pLangHead = (RsrcDirTableHead *)p3;
					p3 += sizeof RsrcDirTableHead;

					Head.Level = 2;
					Head.TypeID  = dwType;
					Head.NameID  = dwName;
					Head.Characteristics = pLangHead->Characteristics;
					Head.TimeDateStamp   = pLangHead->TimeDateStamp;
					Head.VersionMajor    = pLangHead->VersionMajor;
					Head.VersionMinor    = pLangHead->VersionMinor;

					pInfo->vtrTableHeads.push_back(Head);

					for (int LangIndex = 0; LangIndex < pLangHead->NumIDEntries; LangIndex++) {
						RsrcDirEntry *pLangEntry = (RsrcDirEntry *)p3;
						p3 += sizeof(RsrcDirEntry);

						//here can not be subdirs
						wLangID = pLangEntry->IntegerID;

						RsrcDataEntry *pData = (RsrcDataEntry *)(pRsrc + (pLangEntry->DataEntryRVA & ~0x80000000));

						Resource rc;

						rc.NameID     = dwName;
						rc.TypeID     = dwType;
						rc.LanguageID = wLangID;
						rc.DataSize   = pData->Size;
						rc.Codepage   = pData->Codepage;
						rc.Reserved   = pData->Reserved;

						rc.Data = new BYTE[pData->Size];
						memcpy(rc.Data, pRsrc + pData->DataRVA - pInfo->SectionHeader.VirtualAddress, pData->Size);

						pInfo->vtrResources.push_back(rc);
					}
				}
			}
		}
	}

	for (TypeIndex = 0; TypeIndex < pTypeHead->NumIDEntries; TypeIndex++) {	//enumerate resource types
		RsrcDirEntry *pTypeEntry = (RsrcDirEntry *)p;
		p += sizeof(RsrcDirEntry);

		if (pTypeEntry->DataEntryRVA >> 31) {	//pTypeEntry->DataEntryRVA & 0x80000000
			dwType = pTypeEntry->IntegerID;

			char *p2 = pRsrc + (pTypeEntry->SubdirectoryRVA & ~0x80000000);
			RsrcDirTableHead *pNameHead = (RsrcDirTableHead *)p2;
			p2 += sizeof RsrcDirTableHead;

			Head.Level   = 1;
			Head.TypeID  = dwType;
			Head.Characteristics = pNameHead->Characteristics;
			Head.TimeDateStamp   = pNameHead->TimeDateStamp;
			Head.VersionMajor    = pNameHead->VersionMajor;
			Head.VersionMinor    = pNameHead->VersionMinor;

			pInfo->vtrTableHeads.push_back(Head);

			for (int NameIndex = 0; NameIndex < pNameHead->NumNameEntries; NameIndex++) {
				RsrcDirEntry *pNameEntry = (RsrcDirEntry *)p2;
				p2 += sizeof(RsrcDirEntry);

				if (pNameEntry->DataEntryRVA >> 31) {
					dwName = (DWORD)GetRsrcString(pInfo, pRsrc, pNameEntry->NameRVA & ~(1<<31));

					char *p3 = pRsrc + (pNameEntry->SubdirectoryRVA & ~0x80000000);
					RsrcDirTableHead *pLangHead = (RsrcDirTableHead *)p3;
					p3 += sizeof RsrcDirTableHead;

					Head.Level   = 2;
					Head.TypeID  = dwType;
					Head.NameID  = dwName;
					Head.Characteristics = pLangHead->Characteristics;
					Head.TimeDateStamp   = pLangHead->TimeDateStamp;
					Head.VersionMajor    = pLangHead->VersionMajor;
					Head.VersionMinor    = pLangHead->VersionMinor;

					pInfo->vtrTableHeads.push_back(Head);

					for (int LangIndex = 0; LangIndex < pLangHead->NumIDEntries; LangIndex++) {
						RsrcDirEntry *pLangEntry = (RsrcDirEntry *)p3;
						p3 += sizeof(RsrcDirEntry);

						//here can not be subdirs
						wLangID = pLangEntry->IntegerID;

						RsrcDataEntry *pData = (RsrcDataEntry *)(pRsrc + (pLangEntry->DataEntryRVA & ~0x80000000));

						Resource rc;

						rc.NameID     = dwName;
						rc.TypeID     = dwType;
						rc.LanguageID = wLangID;
						rc.DataSize   = pData->Size;
						rc.Codepage   = pData->Codepage;
						rc.Reserved   = pData->Reserved;

						rc.Data = new BYTE[pData->Size];
						memcpy(rc.Data, pRsrc + pData->DataRVA - pInfo->SectionHeader.VirtualAddress, pData->Size);

						pInfo->vtrResources.push_back(rc);
					}
				}
			}

			for (NameIndex = 0; NameIndex < pNameHead->NumIDEntries; NameIndex++) {
				RsrcDirEntry *pNameEntry = (RsrcDirEntry *)p2;
				p2 += sizeof(RsrcDirEntry);

				if (pNameEntry->DataEntryRVA >> 31) {
					dwName = pNameEntry->IntegerID;

					char *p3 = pRsrc + (pNameEntry->SubdirectoryRVA & ~0x80000000);
					RsrcDirTableHead *pLangHead = (RsrcDirTableHead *)p3;
					p3 += sizeof RsrcDirTableHead;

					Head.Level   = 2;
					Head.TypeID  = dwType;
					Head.NameID  = dwName;
					Head.Characteristics = pLangHead->Characteristics;
					Head.TimeDateStamp   = pLangHead->TimeDateStamp;
					Head.VersionMajor    = pLangHead->VersionMajor;
					Head.VersionMinor    = pLangHead->VersionMinor;

					pInfo->vtrTableHeads.push_back(Head);

					for (int LangIndex = 0; LangIndex < pLangHead->NumIDEntries; LangIndex++) {
						RsrcDirEntry *pLangEntry = (RsrcDirEntry *)p3;
						p3 += sizeof(RsrcDirEntry);

						wLangID = pLangEntry->IntegerID;

						//here can not be subdirs

						RsrcDataEntry *pData = (RsrcDataEntry *)(pRsrc + (pLangEntry->DataEntryRVA & ~0x80000000));

						Resource rc;

						rc.NameID     = dwName;
						rc.TypeID     = dwType;
						rc.LanguageID = wLangID;
						rc.DataSize   = pData->Size;
						rc.Codepage   = pData->Codepage;
						rc.Reserved   = pData->Reserved;

						rc.Data = new BYTE[pData->Size];
						memcpy(rc.Data, pRsrc + pData->DataRVA - pInfo->SectionHeader.VirtualAddress, pData->Size);

						pInfo->vtrResources.push_back(rc);
					}
				}
			}
		}
	}
}

HANDLE _BeginUpdateResourceA(LPCSTR lpFileName, BOOL bDelete) {
	RsrcInfo *pInfo = new RsrcInfo;

	if ((pInfo->File = fopen(lpFileName, "r+b")) == NULL) {
		SetLastError(ERROR_OPEN_FAILED);
		delete pInfo;
		return NULL;
	}

	//seek to the beginning of the PE header
	fseek(pInfo->File, 0x3c, SEEK_SET);
	DWORD dw;
	fread(&dw, 4, 1, pInfo->File);

	fseek(pInfo->File, dw, SEEK_SET);
	fread(&dw, 4, 1, pInfo->File);
	if (dw != 0x00004550) {
		SetLastError(ERROR_RESOURCE_DATA_NOT_FOUND);	//a non-PE file does not contain resources
		fclose(pInfo->File);
		delete pInfo;
		return NULL;
	}

	pInfo->FilePtrToPEHdr = ftell(pInfo->File);

	//read in the file header
	fread(&pInfo->PEHeader, sizeof(IMAGE_FILE_HEADER), 1, pInfo->File);

	//skip the optional header
	fseek(pInfo->File, pInfo->PEHeader.SizeOfOptionalHeader, SEEK_CUR);

	//locate the '.rsrc' section
	int RsrcSize = 0;
	for (int i = 0; i < pInfo->PEHeader.NumberOfSections; i++) {
		fread(&pInfo->SectionHeader, sizeof(IMAGE_SECTION_HEADER), 1, pInfo->File);

		if (!strnicmp((char *)pInfo->SectionHeader.Name, ".rsrc", 8)) {
			fseek(pInfo->File, pInfo->SectionHeader.PointerToRawData, SEEK_SET);
			break;
		}
	}

	if (pInfo->SectionHeader.SizeOfRawData == 0) {
		SetLastError(ERROR_RESOURCE_DATA_NOT_FOUND);
		fclose(pInfo->File);
		delete pInfo;
		return NULL;
	}

	if (!bDelete) {	//deleting resources are implementing by not adding them to the table
		//read in the resource table into memory
		char *pRsrc = new char[pInfo->SectionHeader.SizeOfRawData];
		fread(pRsrc, 1, pInfo->SectionHeader.SizeOfRawData, pInfo->File);

		FillRsrcEntries(pInfo, pRsrc);
		delete[] pRsrc;
	}

	return (HANDLE)pInfo;
}

HANDLE _BeginUpdateResourceW(LPCWSTR lpFileName, BOOL bDelete) {
	int nLen = WideCharToMultiByte(CP_ACP, 0, lpFileName, -1, NULL, NULL, NULL, 0);
	LPSTR lpsFileName = new char[nLen];
	WideCharToMultiByte(CP_ACP, 0, lpFileName, -1, lpsFileName, nLen, NULL, 0);

	HANDLE Result = _BeginUpdateResourceA(lpsFileName, bDelete);

	delete[] lpsFileName;

	return Result;
}

////////////////////////////////////////////////////////////////////////
// _UpdateResource()

BOOL _UpdateResourceA(HANDLE hUpdateResource, LPCSTR lpType, LPCSTR lpName, WORD wLanguage, LPVOID lpData, DWORD cbDataSize) {
	LPWSTR lpwType = (LPWSTR)lpType, lpwName = (LPWSTR)lpName;	//will be changed later if strings

	if (HIWORD(lpType) != 0) {
		int nLen = MultiByteToWideChar(CP_ACP, 0, lpType, -1, NULL, NULL);
		lpwType = new WCHAR[nLen];
		MultiByteToWideChar(CP_ACP, 0, lpType, -1, lpwType, nLen);
	}
	if (HIWORD(lpName) != 0) {
		int nLen = MultiByteToWideChar(CP_ACP, 0, lpName, -1, NULL, NULL);
		lpwName = new WCHAR[nLen];
		MultiByteToWideChar(CP_ACP, 0, lpName, -1, lpwName, nLen);
	}

	BOOL RetCode = _UpdateResourceW(hUpdateResource, lpwType, lpwName, wLanguage, lpData, cbDataSize);

	if (HIWORD(lpType) != 0) {
		delete[] lpwType;
	}
	if (HIWORD(lpName) != 0) {
		delete[] lpwName;
	}

	return RetCode;
}

BOOL _UpdateResourceW(HANDLE hUpdateResource, LPCWSTR lpType, LPCWSTR lpName, WORD wLanguage, LPVOID lpData, DWORD cbDataSize) {
	if (!hUpdateResource) {
		SetLastError(ERROR_INVALID_HANDLE);
		return FALSE;
	}

	RsrcInfo *pRsrcInfo = (RsrcInfo *)hUpdateResource;

	Resource rc;

	if (HIWORD(lpType) == 0) {
		rc.TypeID = (DWORD)lpType;
	}
	else {
		rc.TypeString = NULL;
		for (vector<LPWSTR>::iterator I = pRsrcInfo->vtrStrings.begin(); I != pRsrcInfo->vtrStrings.end(); I++) {
			if (wcsicmp(lpType, *I) == 0) {
				rc.TypeString = *I;
				break;
			}
		}

		if (!rc.TypeString) {
			rc.TypeString = new WCHAR[wcslen(lpType) + 1];
			wcscpy(rc.TypeString, lpType);
			pRsrcInfo->vtrStrings.push_back(rc.TypeString);
		}
	}

	if (HIWORD(lpName) == 0) {
		rc.NameID = (DWORD)lpName;
	}
	else {
		rc.NameString = NULL;
		for (vector<LPWSTR>::iterator I = pRsrcInfo->vtrStrings.begin(); I != pRsrcInfo->vtrStrings.end(); I++) {
			if (wcsicmp(lpName, *I) == 0) {
				rc.NameString = *I;
				break;
			}
		}

		if (!rc.NameString) {
			rc.NameString = new WCHAR[wcslen(lpName) + 1];
			wcscpy(rc.NameString, lpName);
			pRsrcInfo->vtrStrings.push_back(rc.NameString);
		}
	}

	rc.LanguageID = wLanguage;
	rc.Codepage   = 0;
	rc.Reserved   = 0;
	rc.DataSize   = cbDataSize;

	if (lpData) {
		rc.Data = new BYTE[cbDataSize];
		memcpy(rc.Data, lpData, cbDataSize);
	}

	Resource *pResource = NULL;

	for (vector<Resource>::iterator I = pRsrcInfo->vtrResources.begin(); I != pRsrcInfo->vtrResources.end(); I++) {
		if (*I == rc) {
			pResource = I;
			break;
		}
	}

	if (pResource) {
		delete[] pResource->Data;

		if (lpData)
			*pResource = rc;
		else
			pRsrcInfo->vtrResources.erase(pResource);
	}
	else {
		if (lpData)
			pRsrcInfo->vtrResources.push_back(rc);
	}

	return TRUE;
}

////////////////////////////////////////////////////////////////////////
// _GetRsrcData()

LPBYTE _GetRsrcDataA(HANDLE hUpdateResource, LPCSTR  lpType, LPCSTR  lpName, WORD wLanguage, int *pDataSize) {
	LPWSTR lpwType = (LPWSTR)lpType, lpwName = (LPWSTR)lpName;	//will be changed later if strings

	if (HIWORD(lpType) != 0) {
		int nLen = MultiByteToWideChar(CP_ACP, 0, lpType, -1, NULL, NULL);
		lpwType = new WCHAR[nLen];
		MultiByteToWideChar(CP_ACP, 0, lpType, -1, lpwType, nLen);
	}
	if (HIWORD(lpName) != 0) {
		int nLen = MultiByteToWideChar(CP_ACP, 0, lpName, -1, NULL, NULL);
		lpwName = new WCHAR[nLen];
		MultiByteToWideChar(CP_ACP, 0, lpName, -1, lpwName, nLen);
	}

	LPBYTE Ptr = _GetRsrcDataW(hUpdateResource, lpwType, lpwName, wLanguage, pDataSize);

	if (HIWORD(lpType) != 0) {
		delete[] lpwType;
	}
	if (HIWORD(lpName) != 0) {
		delete[] lpwName;
	}

	return Ptr;
}

LPBYTE _GetRsrcDataW(HANDLE hUpdateResource, LPCWSTR lpType, LPCWSTR lpName, WORD wLanguage, int *pDataSize) {
	if (!hUpdateResource) {
		SetLastError(ERROR_INVALID_HANDLE);
		return FALSE;
	}

	RsrcInfo *pRsrcInfo = (RsrcInfo *)hUpdateResource;

	Resource rc;

	rc.TypeID     = (DWORD)lpType;
	rc.NameID     = (DWORD)lpName;
	rc.LanguageID = wLanguage;

	for (vector<Resource>::iterator I = pRsrcInfo->vtrResources.begin(); I != pRsrcInfo->vtrResources.end(); I++) {
		if (*I == rc) {
			if (pDataSize)
				*pDataSize = I->DataSize;
			return I->Data;
		}
	}

	if (pDataSize)
		*pDataSize = 0;

	SetLastError(ERROR_RESOURCE_NOT_FOUND);

	return NULL;
}

////////////////////////////////////////////////////////////////////////
// _EndUpdateResource()


typedef vector<vector<vector<Resource> > > ResourceTable;

class CmpUnicodeNoCase {
	public:
		bool operator()(const basic_string<wchar_t> &str1, const basic_string<wchar_t> &str2) const {return wcsicmp(str1.c_str(), str2.c_str()) < 0;}
};

#define DWORD_ALIGN(i) (i = (i + 3) & ~3)

static RsrcDirTableHead FindTableHead(const RsrcInfo *pInfo, int Level, DWORD Type = 0, DWORD Name = 0) {
	for (vector<TableHead>::const_iterator I = pInfo->vtrTableHeads.begin(); I != pInfo->vtrTableHeads.end(); I++) {
		if (I->Level == Level) {
			switch (Level) {
				case 0:
					goto FoundIt;
				case 1:
					if (RsrcIDEqual(Type, I->TypeID))
						goto FoundIt;
					break;
				case 2:
					if (RsrcIDEqual(Type, I->TypeID) && RsrcIDEqual(Name, I->NameID))
						goto FoundIt;
					break;
			}
		}
	}
FoundIt:

	RsrcDirTableHead Head;

	if (I == pInfo->vtrTableHeads.end()) {
		Head.Characteristics = 0;
		Head.TimeDateStamp   = time(NULL);
		Head.VersionMajor    = 0;
		Head.VersionMinor    = 0;
	}
	else {
		Head.Characteristics = I->Characteristics;
		Head.TimeDateStamp   = I->TimeDateStamp;
		Head.VersionMajor    = I->VersionMajor;
		Head.VersionMinor    = I->VersionMinor;
	}

	Head.NumNameEntries = 0;
	Head.NumIDEntries   = 0;

	return Head;
}

static void CreateResourceTable(vector<Resource> vtrResources, ResourceTable &Table) {
	int TypeIndex = 0, LastType = 0;	//will add 1 to TypeIndex in first loop

	vector<vector<Resource> > vtrTypesSorted;

	vtrTypesSorted.push_back(vector<Resource>());
	vtrTypesSorted[TypeIndex].push_back(vtrResources.back());
	LastType = vtrResources.back().TypeID;

	vtrResources.pop_back();

	while (!vtrResources.empty()) {
		if (RsrcIDEqual(vtrResources.back().TypeID, LastType))
			vtrTypesSorted[TypeIndex].push_back(vtrResources.back());
		else {
			vtrTypesSorted.push_back(vector<Resource>());
			TypeIndex++;
			vtrTypesSorted[TypeIndex].push_back(vtrResources.back());
			LastType = vtrResources.back().TypeID;
		}

		vtrResources.pop_back();
	}

	Table.assign(vtrTypesSorted.size());

	for (TypeIndex = 0; TypeIndex < vtrTypesSorted.size(); TypeIndex++) {
		int NameIndex = 0, LastName = 0;

		Table[TypeIndex].push_back(vector<Resource>());
		Table[TypeIndex][NameIndex].push_back(vtrTypesSorted[TypeIndex].back());
		LastName = vtrTypesSorted[TypeIndex].back().NameID;

		vtrTypesSorted[TypeIndex].pop_back();

		while (!vtrTypesSorted[TypeIndex].empty()) {
			if (RsrcIDEqual(vtrTypesSorted[TypeIndex].back().NameID, LastName))
				Table[TypeIndex][NameIndex].push_back(vtrTypesSorted[TypeIndex].back());
			else {
				Table[TypeIndex].push_back(vector<Resource>());
				NameIndex++;
				Table[TypeIndex][NameIndex].push_back(vtrTypesSorted[TypeIndex].back());
				LastName = vtrTypesSorted[TypeIndex].back().NameID;
			}

			vtrTypesSorted[TypeIndex].pop_back();
		}
	}

	reverse(Table.begin(), Table.end());
}

static int CalcDirTableSize(const ResourceTable &Table) {
	int Size = sizeof(RsrcDirTableHead) + Table.size() * (sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));	//1 directory entry + 1 table head for the next level

	for (int i = 0; i < Table.size(); i++) {
		Size += Table[i].size() * (sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));

		for (int j = 0; j < Table[i].size(); j++)
			Size += Table[i][j].size() * (sizeof(RsrcDirEntry) + sizeof(RsrcDataEntry));
	}

	return Size;
}

static int CalcStringTableSize(const RsrcInfo *pInfo) {
	int Size = 0;

	for (vector<LPWSTR>::const_iterator I = pInfo->vtrStrings.begin(); I != pInfo->vtrStrings.end(); I++)
		Size += (wcslen(*I) + 1) * 2;

	return Size;
}

static int CalcRawDataSize(const ResourceTable &Table) {
	int Size = 0;

	for (int i = 0; i < Table.size(); i++) {
		for (int j = 0; j < Table[i].size(); j++) {
			for (int k = 0; k < Table[i][j].size(); k++) {
				Size += Table[i][j][k].DataSize;
				DWORD_ALIGN(Size);
			}
		}
	}

	return Size;
}

static int CalcNumDirEntries(const ResourceTable &Table, bool fID, int Level, int Type = 0, int Name = 0) {
	int Count = 0;

	switch (Level) {
		case 0: {
			for (int i = 0; i < Table.size(); i++) {
				if ((fID && !HIWORD(Table[i][0][0].TypeID)) || (!fID && HIWORD(Table[i][0][0].TypeID)))
					Count++;
			}

			break;
		}
		case 1: {
			for (int i = 0; i < Table[Type].size(); i++) {
				if ((fID && !HIWORD(Table[Type][i][0].NameID)) || (!fID && HIWORD(Table[Type][i][0].NameID)))
					Count++;
			}

			break;
		}
		case 2: {
			Count = Table[Type][Name].size();

			break;
		}
	}

	return Count;
}

static map<basic_string<wchar_t>, int, CmpUnicodeNoCase> GenerateStringTable(const RsrcInfo *pInfo, char *pStringTable) {
	WORD *StringTable = (WORD *)pStringTable;
	int Offset = 0;
	map<basic_string<wchar_t>, int, CmpUnicodeNoCase> Relocs;

	for (vector<LPWSTR>::const_iterator I = pInfo->vtrStrings.begin(); I != pInfo->vtrStrings.end(); I++) {
		Relocs[*I] = Offset * 2;	//StringTable is words, raw data is bytes
		WORD len = wcslen(*I);
		StringTable[Offset++] = len;
		memcpy(&StringTable[Offset], *I, len * 2);
		Offset += len;
	}

	return Relocs;
}

static void GenerateCode(const RsrcInfo *pInfo, const ResourceTable &Table, char *pRsrc, map<basic_string<wchar_t>, int, CmpUnicodeNoCase> StringTable, int StringTableOffset, int RawDataOffset) {
	int TableAddress = 0, TypeHeadAddress = 0, DataAddress = 0;
	int RsrcIndex = 1;

	RsrcDirTableHead TypeHead = FindTableHead(pInfo, 0);

	TypeHead.NumNameEntries   = CalcNumDirEntries(Table, false, 0);
	TypeHead.NumIDEntries     = CalcNumDirEntries(Table, true,  0);

	*(RsrcDirTableHead *)pRsrc = TypeHead;

	TableAddress += sizeof(RsrcDirTableHead) + (TypeHead.NumNameEntries + TypeHead.NumIDEntries) * sizeof(RsrcDirEntry);

	for (int Type = 0; Type < TypeHead.NumNameEntries; Type++) {
		//create the subdirectory table
		int NameHeadAddress = TableAddress;

		RsrcDirTableHead NameHead = FindTableHead(pInfo, 1, Table[Type][0][0].TypeID);

		NameHead.NumNameEntries   = CalcNumDirEntries(Table, false, 1, Type);
		NameHead.NumIDEntries     = CalcNumDirEntries(Table, true,  1, Type);

		*(RsrcDirTableHead *)&pRsrc[NameHeadAddress] = NameHead;

		TableAddress += sizeof(RsrcDirTableHead) + (NameHead.NumNameEntries + NameHead.NumIDEntries) * sizeof(RsrcDirEntry);

		for (int Name = 0; Name < NameHead.NumNameEntries; Name++) {
			//create the subdirectory table
			int LangHeadAddress = TableAddress;

			RsrcDirTableHead LangHead = FindTableHead(pInfo, 2, Table[Type][0][0].TypeID, Table[Type][Name][0].NameID);

			LangHead.NumNameEntries   = 0;
			LangHead.NumIDEntries     = CalcNumDirEntries(Table, true, 2, Type, Name);

			*(RsrcDirTableHead *)&pRsrc[LangHeadAddress] = LangHead;

			TableAddress += sizeof(RsrcDirTableHead) + (LangHead.NumNameEntries + LangHead.NumIDEntries) * sizeof(RsrcDirEntry);

			for (int Lang = 0; Lang < LangHead.NumIDEntries; Lang++) {
				//create data entry
				RsrcDataEntry *pDataEntry = (RsrcDataEntry *)(pRsrc + TableAddress);
				TableAddress += sizeof(RsrcDataEntry);

				const Resource &rc = Table[Type][Name][Lang];

				memcpy(pRsrc + RawDataOffset + DataAddress, rc.Data, rc.DataSize);
				pDataEntry->DataRVA  = RawDataOffset + DataAddress + pInfo->SectionHeader.VirtualAddress;
				pDataEntry->Size     = rc.DataSize;
				pDataEntry->Codepage = rc.Codepage;
				pDataEntry->Reserved = rc.Reserved;

				RsrcIndex++;

				DataAddress += rc.DataSize;
				DWORD_ALIGN(DataAddress);

				//create the dir entry
				RsrcDirEntry *pEntry = (RsrcDirEntry *)(pRsrc + LangHeadAddress + Lang * sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));
				pEntry->IntegerID    = Table[Type][Name][Lang].LanguageID;
				pEntry->DataEntryRVA = TableAddress - sizeof(RsrcDataEntry);
			}

			//create the dir entry
			RsrcDirEntry *pEntry = (RsrcDirEntry *)(pRsrc + NameHeadAddress + Name * sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));
			pEntry->NameRVA = (StringTable[Table[Type][Name][0].NameString] + StringTableOffset) | (1<<31);
			pEntry->SubdirectoryRVA = LangHeadAddress | (1<<31);
		}

		for (; Name < NameHead.NumNameEntries + NameHead.NumIDEntries; Name++) {
			//create the subdirectory table
			int LangHeadAddress = TableAddress;

			RsrcDirTableHead LangHead = FindTableHead(pInfo, 2, Table[Type][0][0].TypeID, Table[Type][Name][0].NameID);

			LangHead.NumNameEntries   = 0;
			LangHead.NumIDEntries     = CalcNumDirEntries(Table, true, 2, Type, Name);

			*(RsrcDirTableHead *)&pRsrc[LangHeadAddress] = LangHead;

			TableAddress += sizeof(RsrcDirTableHead) + (LangHead.NumNameEntries + LangHead.NumIDEntries) * sizeof(RsrcDirEntry);

			for (int Lang = 0; Lang < LangHead.NumIDEntries; Lang++) {
				//create data entry
				RsrcDataEntry *pDataEntry = (RsrcDataEntry *)(pRsrc + TableAddress);
				TableAddress += sizeof(RsrcDataEntry);

				const Resource &rc = Table[Type][Name][Lang];

				memcpy(pRsrc + RawDataOffset + DataAddress, rc.Data, rc.DataSize);
				pDataEntry->DataRVA  = RawDataOffset + DataAddress + pInfo->SectionHeader.VirtualAddress;
				pDataEntry->Size     = rc.DataSize;
				pDataEntry->Codepage = rc.Codepage;
				pDataEntry->Reserved = rc.Reserved;

				RsrcIndex++;

				DataAddress += rc.DataSize;
				DWORD_ALIGN(DataAddress);

				//create the dir entry
				RsrcDirEntry *pEntry = (RsrcDirEntry *)(pRsrc + LangHeadAddress + Lang * sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));
				pEntry->IntegerID    = Table[Type][Name][Lang].LanguageID;
				pEntry->DataEntryRVA = TableAddress - sizeof(RsrcDataEntry);
			}

			//create the dir entry
			RsrcDirEntry *pEntry = (RsrcDirEntry *)(pRsrc + NameHeadAddress + Name * sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));
			pEntry->IntegerID = Table[Type][Name][0].NameID;
			pEntry->SubdirectoryRVA = LangHeadAddress | (1<<31);
		}

		//create the dir entry
		RsrcDirEntry *pEntry = (RsrcDirEntry *)(pRsrc + TypeHeadAddress + Type * sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));
		pEntry->NameRVA = (StringTable[Table[Type][0][0].TypeString] + StringTableOffset) | (1<<31);
		pEntry->SubdirectoryRVA = NameHeadAddress | (1<<31);
	}

	for (; Type < TypeHead.NumNameEntries + TypeHead.NumIDEntries; Type++) {
		//create the subdirectory table
		int NameHeadAddress = TableAddress;

		RsrcDirTableHead NameHead = FindTableHead(pInfo, 1, Table[Type][0][0].TypeID);

		NameHead.NumNameEntries   = CalcNumDirEntries(Table, false, 1, Type);
		NameHead.NumIDEntries     = CalcNumDirEntries(Table, true,  1, Type);

		*(RsrcDirTableHead *)&pRsrc[NameHeadAddress] = NameHead;

		TableAddress += sizeof(RsrcDirTableHead) + (NameHead.NumNameEntries + NameHead.NumIDEntries) * sizeof(RsrcDirEntry);

		for (int Name = 0; Name < NameHead.NumNameEntries; Name++) {
			//create the subdirectory table
			int LangHeadAddress = TableAddress;

			RsrcDirTableHead LangHead = FindTableHead(pInfo, 2, Table[Type][0][0].TypeID, Table[Type][Name][0].NameID);

			LangHead.NumNameEntries   = 0;
			LangHead.NumIDEntries     = CalcNumDirEntries(Table, true, 2, Type, Name);

			*(RsrcDirTableHead *)&pRsrc[LangHeadAddress] = LangHead;

			TableAddress += sizeof(RsrcDirTableHead) + (LangHead.NumNameEntries + LangHead.NumIDEntries) * sizeof(RsrcDirEntry);

			for (int Lang = 0; Lang < LangHead.NumIDEntries; Lang++) {
				//create data entry
				RsrcDataEntry *pDataEntry = (RsrcDataEntry *)(pRsrc + TableAddress);
				TableAddress += sizeof(RsrcDataEntry);

				const Resource &rc = Table[Type][Name][Lang];

				memcpy(pRsrc + RawDataOffset + DataAddress, rc.Data, rc.DataSize);
				pDataEntry->DataRVA  = RawDataOffset + DataAddress + pInfo->SectionHeader.VirtualAddress;
				pDataEntry->Size     = rc.DataSize;
				pDataEntry->Codepage = rc.Codepage;
				pDataEntry->Reserved = rc.Reserved;

				RsrcIndex++;

				DataAddress += rc.DataSize;
				DWORD_ALIGN(DataAddress);

				//create the dir entry
				RsrcDirEntry *pEntry = (RsrcDirEntry *)(pRsrc + LangHeadAddress + Lang * sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));
				pEntry->IntegerID    = Table[Type][Name][Lang].LanguageID;
				pEntry->DataEntryRVA = TableAddress - sizeof(RsrcDataEntry);
			}

			//create the dir entry
			RsrcDirEntry *pEntry = (RsrcDirEntry *)(pRsrc + NameHeadAddress + Name * sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));
			pEntry->NameRVA = (StringTable[Table[Type][Name][0].NameString] + StringTableOffset) | (1<<31);
			pEntry->SubdirectoryRVA = LangHeadAddress | (1<<31);
		}

		for (; Name < NameHead.NumNameEntries + NameHead.NumIDEntries; Name++) {
			//create the subdirectory table
			int LangHeadAddress = TableAddress;

			RsrcDirTableHead LangHead = FindTableHead(pInfo, 2, Table[Type][0][0].TypeID, Table[Type][Name][0].NameID);

			LangHead.NumNameEntries   = 0;
			LangHead.NumIDEntries     = CalcNumDirEntries(Table, true, 2, Type, Name);

			*(RsrcDirTableHead *)&pRsrc[LangHeadAddress] = LangHead;

			TableAddress += sizeof(RsrcDirTableHead) + (LangHead.NumNameEntries + LangHead.NumIDEntries) * sizeof(RsrcDirEntry);

			for (int Lang = 0; Lang < LangHead.NumIDEntries; Lang++) {
				//create data entry
				RsrcDataEntry *pDataEntry = (RsrcDataEntry *)(pRsrc + TableAddress);
				TableAddress += sizeof(RsrcDataEntry);

				const Resource &rc = Table[Type][Name][Lang];

				memcpy(pRsrc + RawDataOffset + DataAddress, rc.Data, rc.DataSize);
				pDataEntry->DataRVA  = RawDataOffset + DataAddress + pInfo->SectionHeader.VirtualAddress;
				pDataEntry->Size     = rc.DataSize;
				pDataEntry->Codepage = rc.Codepage;
				pDataEntry->Reserved = rc.Reserved;

				RsrcIndex++;

				DataAddress += rc.DataSize;
				DWORD_ALIGN(DataAddress);

				//create the dir entry
				RsrcDirEntry *pEntry = (RsrcDirEntry *)(pRsrc + LangHeadAddress + Lang * sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));
				pEntry->IntegerID    = Table[Type][Name][Lang].LanguageID;
				pEntry->DataEntryRVA = TableAddress - sizeof(RsrcDataEntry);
			}

			//create the dir entry
			RsrcDirEntry *pEntry = (RsrcDirEntry *)(pRsrc + NameHeadAddress + Name * sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));
			pEntry->IntegerID = Table[Type][Name][0].NameID;
			pEntry->SubdirectoryRVA = LangHeadAddress | (1<<31);
		}

		//create the dir entry
		RsrcDirEntry *pEntry = (RsrcDirEntry *)(pRsrc + TypeHeadAddress + Type * sizeof(RsrcDirEntry) + sizeof(RsrcDirTableHead));
		pEntry->IntegerID = Table[Type][0][0].TypeID;
		pEntry->SubdirectoryRVA = NameHeadAddress | (1<<31);
	}
}

void RemoveUnreferencedStrings(RsrcInfo *pInfo) {
	for (vector<LPWSTR>::iterator StringIter = pInfo->vtrStrings.begin(); StringIter != pInfo->vtrStrings.end(); StringIter++) {
		bool Referenced = false;
		
		for (vector<Resource>::iterator RsrcIter = pInfo->vtrResources.begin(); RsrcIter != pInfo->vtrResources.end(); RsrcIter++) {
			if (RsrcIter->TypeString == *StringIter || RsrcIter->NameString == *StringIter) {
				Referenced = true;
				break;
			}
		}

		if (!Referenced) {
			delete[] *StringIter;
			pInfo->vtrStrings.erase(StringIter);
		}
	}
}

static void CleanupRsrcInfo(RsrcInfo *pInfo) {
	fclose(pInfo->File);

	for (vector<Resource>::iterator I = pInfo->vtrResources.begin(); I != pInfo->vtrResources.end(); I++)
		delete[] I->Data;

	for (vector<LPWSTR>::iterator StringIter = pInfo->vtrStrings.begin(); StringIter != pInfo->vtrStrings.end(); StringIter++)
		delete[] *StringIter;

	delete pInfo;
}

static int RelocateSections(RsrcInfo *pInfo, int NewRsrcSize, int OldImageSize, int FileAlignment, int SectionAlignment) {
	IMAGE_SECTION_HEADER sh;
	int AlignedImageSize;

	int AlignedRsrcSize = (NewRsrcSize + FileAlignment - 1) & ~(FileAlignment - 1);

	fseek(pInfo->File, pInfo->FilePtrToPEHdr + sizeof(IMAGE_FILE_HEADER) + pInfo->PEHeader.SizeOfOptionalHeader, SEEK_SET);

	if (NewRsrcSize < pInfo->SectionHeader.SizeOfRawData) {
		//we don't have to relocate anything - just change the size
		for (int Section = 0; Section < pInfo->PEHeader.NumberOfSections; Section++) {
			fread(&sh, sizeof(IMAGE_SECTION_HEADER), 1, pInfo->File);

			if (strnicmp((char *)sh.Name, ".rsrc", 8) == 0) {
				sh.SizeOfRawData    = AlignedRsrcSize;
				sh.Misc.VirtualSize = NewRsrcSize;
				fseek(pInfo->File, -sizeof(IMAGE_SECTION_HEADER), SEEK_CUR);
				fwrite(&sh, sizeof(IMAGE_SECTION_HEADER), 1, pInfo->File);
				AlignedImageSize = OldImageSize - sh.Misc.VirtualSize + NewRsrcSize;
				AlignedImageSize = (AlignedImageSize + SectionAlignment - 1) & ~(SectionAlignment - 1);

				break;
			}
		}

		fseek(pInfo->File, pInfo->FilePtrToPEHdr + sizeof(IMAGE_FILE_HEADER) + offsetof(IMAGE_OPTIONAL_HEADER, DataDirectory[2]) + 4, SEEK_SET);
		fwrite(&NewRsrcSize, 4, 1, pInfo->File);

		fseek(pInfo->File, pInfo->FilePtrToPEHdr + sizeof(IMAGE_FILE_HEADER) + offsetof(IMAGE_OPTIONAL_HEADER, SizeOfImage), SEEK_SET);
		fwrite(&AlignedImageSize, 4, 1, pInfo->File);

		return ERROR_SUCCESS;
	}

	int FilePtrToRsrc = 0;

	for (int Section = 0; Section < pInfo->PEHeader.NumberOfSections; Section++) {
		fread(&sh, sizeof(IMAGE_SECTION_HEADER), 1, pInfo->File);

		if (sh.VirtualAddress > pInfo->SectionHeader.VirtualAddress || sh.PointerToRawData > pInfo->SectionHeader.PointerToRawData)
			return ERROR_CANT_MOVE_SECTION;

		if (!FilePtrToRsrc && strnicmp((char *)sh.Name, ".rsrc", 8) == 0) {
			FilePtrToRsrc = ftell(pInfo->File) - sizeof(IMAGE_SECTION_HEADER);
			AlignedImageSize = OldImageSize - sh.Misc.VirtualSize + NewRsrcSize;
			AlignedImageSize = (AlignedImageSize + SectionAlignment - 1) & ~(SectionAlignment - 1);
		}
	}

	fseek(pInfo->File, FilePtrToRsrc, SEEK_SET);

	pInfo->SectionHeader.SizeOfRawData    = AlignedRsrcSize;
	pInfo->SectionHeader.Misc.VirtualSize = NewRsrcSize;

	fwrite(&pInfo->SectionHeader, sizeof(IMAGE_SECTION_HEADER), 1, pInfo->File);

	fseek(pInfo->File, pInfo->FilePtrToPEHdr + sizeof(IMAGE_FILE_HEADER) + offsetof(IMAGE_OPTIONAL_HEADER, DataDirectory[2]) + 4, SEEK_SET);
	fwrite(&NewRsrcSize, 4, 1, pInfo->File);

	fseek(pInfo->File, pInfo->FilePtrToPEHdr + sizeof(IMAGE_FILE_HEADER) + offsetof(IMAGE_OPTIONAL_HEADER, SizeOfImage), SEEK_SET);
	fwrite(&AlignedImageSize, 4, 1, pInfo->File);

/*
	for (int Section = 0; Section < pInfo->PEHeader.NumberOfSections; Section++) {
		if ((sh.VirtualAddress > pInfo->SectionHeader.VirtualAddress) &&
		    strnicmp((char *)sh.Name, ".rsrc", 8) && strnicmp((char *)sh.Name, ".reloc", 8))
			//for now we can only relocate the ".reloc" section
			return ERROR_CANT_MOVE_SECTION;
	}

	int FilePtrToRsrc = 0;

	for (Section = 0; Section < pInfo->PEHeader.NumberOfSections; Section++) {
		fread(&sh, sizeof(IMAGE_SECTION_HEADER), 1, pInfo->File);

		if (!FilePtrToRsrc && (strnicmp(sh.Name, ".rsrc", 8) == 0)) {
			FilePtrToRsrc = ftell(pInfo->File) - sizeof(IMAGE_SECTION_HEADER);
		}
		else {
			if (sh.VirtualAddress > pInfo->SectionHeader.VirtualAddress) {
				if (!strnicmp((char *)sh.Name, ".reloc", 8)) {
					sh.VirtualAddress = pInfo->SectionHeader.VirtualAddress;
					pInfo->SectionHeader.VirtualAddress += ((sh.VirtualSize + 0x1000 - 1) & ~(0x1000-1);
				}

				fseek(pInfo->File, -sizeof(IMAGE_SECTION_HEADER), SEEK_CUR);
				fwrite(&sh, sizeof(IMAGE_SECTION_HEADER), 1, pInfo->File);
			}
		}
	}
*/
	return ERROR_SUCCESS;
}

BOOL _EndUpdateResource(HANDLE hUpdateResource, BOOL fDiscard) {
	if (!hUpdateResource) {
		SetLastError(ERROR_INVALID_HANDLE);
		return FALSE;
	}

	RsrcInfo *pRsrcInfo = (RsrcInfo *)hUpdateResource;

	if (fDiscard) {
		CleanupRsrcInfo(pRsrcInfo);
		return TRUE;
	}

	sort(pRsrcInfo->vtrResources.begin(), pRsrcInfo->vtrResources.end());

	RemoveUnreferencedStrings(pRsrcInfo);

	ResourceTable Table;
	CreateResourceTable(pRsrcInfo->vtrResources, Table);

	IMAGE_OPTIONAL_HEADER oh;
	fseek(pRsrcInfo->File, pRsrcInfo->FilePtrToPEHdr + sizeof(IMAGE_FILE_HEADER), SEEK_SET);
	fread(&oh, sizeof(IMAGE_OPTIONAL_HEADER), 1, pRsrcInfo->File);

	//we do 2 passes: first we calculate all sizes and put the '.rsrc' section last in file and Virtual memory
	int DirTableSize = CalcDirTableSize(Table), StringTableSize = CalcStringTableSize(pRsrcInfo), RawDataSize = CalcRawDataSize(Table);

	DWORD_ALIGN(DirTableSize);
	DWORD_ALIGN(StringTableSize);

	DWORD TotalSize = DirTableSize + StringTableSize + RawDataSize, AlignedSize = (TotalSize + oh.FileAlignment - 1) & ~(oh.FileAlignment - 1);

	int Error = RelocateSections(pRsrcInfo, TotalSize, oh.SizeOfImage, oh.FileAlignment, oh.SectionAlignment);
	if (Error != ERROR_SUCCESS) {
		SetLastError(Error);
		CleanupRsrcInfo(pRsrcInfo);
		return FALSE;
	}

	char *pRsrc = new char[AlignedSize];
	memset(pRsrc, 0, AlignedSize);

	//...then we generate the code
	map<basic_string<wchar_t>, int, CmpUnicodeNoCase> StringRelocs = GenerateStringTable(pRsrcInfo, pRsrc + DirTableSize);
	GenerateCode(pRsrcInfo, Table, pRsrc, StringRelocs, DirTableSize, DirTableSize + StringTableSize);

	//write to file
	fseek(pRsrcInfo->File, pRsrcInfo->SectionHeader.PointerToRawData, SEEK_SET);
	fwrite(pRsrc, 1, AlignedSize, pRsrcInfo->File);

	//done - clean up
	delete[] pRsrc;

	CleanupRsrcInfo(pRsrcInfo);

	return TRUE;
}